@@ -2,6 +2,7 @@ import glob
 import io
 import os
 import pickle
+import shutil
 import tempfile
 from contextlib import contextmanager
 from io import BytesIO
@@ -19,24 +20,16 @@ ModeType = Literal['r', 'r+', 'w+', 'c']
 MMapModeType = Literal['readonly', 'r', 'copyonwrite', 'c', 'readwrite', 'r+', 'write', 'w+']
 
 
-def get_client(
-        endpoint_url: Optional[str] = os.environ.get('S3_ENDPOINT_URL', None),
-        verify: Optional[bool] = None,
-):
-
-    verify = utils.to_bool(os.environ.get('S3_VERIFY', None)) if verify is None else verify
+def get_client(endpoint_url: Optional[str] = os.getenv('S3_ENDPOINT_URL'), verify: Optional[bool] = None):
+    verify = utils.to_bool(os.getenv('S3_VERIFY')) if verify is None else verify
 
     client = boto3.client('s3', endpoint_url=endpoint_url, verify=verify)
 
     return client
 
 
-def get_s3fs(
-        endpoint_url: Optional[str] = os.environ.get('S3_ENDPOINT_URL', None),
-        verify: Optional[bool] = None,
-) -> S3FileSystem:
-
-    verify = utils.to_bool(os.environ.get('S3_VERIFY', None)) if verify is None else verify
+def get_s3fs(endpoint_url: Optional[str] = os.getenv('S3_ENDPOINT_URL'), verify: Optional[bool] = None) -> S3FileSystem:
+    verify = utils.to_bool(os.getenv('S3_VERIFY')) if verify is None else verify
 
     filesystem = S3FileSystem(client_kwargs={'endpoint_url': endpoint_url, 'verify': verify})
 
@@ -45,8 +38,9 @@ def get_s3fs(
 
 def get_bucket_and_key(path: str) -> Tuple[str, str]:
     url = urlparse(path)
+    bucket, key = url.netloc, url.path.strip('/')
 
-    return url.netloc, url.path.strip('/')
+    return bucket, key
 
 
 def download_s3(bucket: str, key: str, destination: str) -> None:
@@ -120,7 +114,7 @@ def exists(path: str) -> bool:
     return has_keys
 
 
-def generate_path(suffix: Optional[str] = None, prefix: Optional[str] = os.environ.get('DATA_PATH', None)) -> str:
+def generate_path(suffix: Optional[str] = None, prefix: Optional[str] = os.getenv('DATA_PATH')) -> str:
     prefix = prefix if prefix is not None else tempfile.gettempdir()
 
     if not suffix:
@@ -133,49 +127,83 @@ def generate_path(suffix: Optional[str] = None, prefix: Optional[str] = os.envir
 
 def localize_directory(
         path: str,
-        directory: Optional[str] = None,
+        directory: Optional[str] = os.getenv('DATA_EPHEMERAL_PATH'),
         prefix: Optional[str] = None,
         overwrite: bool = True,
 ) -> str:
 
-    if not path.startswith('s3'):
+    target = generate_path() if directory is None else directory
+    target_prefixed = os.path.join(target, prefix) if prefix else target
+
+    if not overwrite and exists(target_prefixed):
+        return target
+
+    elif path.startswith('s3'):
+        cache = generate_path()
+        cache_prefixed = os.path.join(cache, prefix) if prefix else cache
+
+        if overwrite or not exists(cache):
+            bucket, key_base = get_bucket_and_key(path)
+            key_base_prefixed = os.path.join(key_base, prefix) if prefix else key_base
+            key_base_prefixed = key_base_prefixed if key_base_prefixed.endswith('/') else key_base_prefixed + '/'
+
+            for key in _get_s3_keys(bucket, key_base_prefixed):
+                download_s3(bucket, key, generate_path(os.path.relpath(key, key_base), cache))
+
+        if cache != target:
+            shutil.copytree(cache_prefixed, target_prefixed)
+
+        return target
+
+    elif directory is not None:
+        path_prefixed = os.path.join(path, prefix) if prefix else path
+
+        if path != target:
+            shutil.copytree(path_prefixed, target_prefixed)
+
+        return target
+
+    else:
         return path
 
-    directory = generate_path(os.path.basename(path)) if directory is None else directory
 
-    directory_prefixed = os.path.join(directory, prefix) if prefix else directory
-    if not overwrite and os.path.exists(directory_prefixed):
-        return directory
+def localize_file(
+        path: str,
+        directory: Optional[str] = os.getenv('DATA_EPHEMERAL_PATH'),
+        overwrite: bool = True,
+) -> str:
 
-    bucket, key_base = get_bucket_and_key(path)
-    key_base_prefixed = os.path.join(key_base, prefix) if prefix else key_base
-    key_base_prefixed = key_base_prefixed if key_base_prefixed.endswith('/') else key_base_prefixed + '/'
+    basename = os.path.basename(path)
+    target = generate_path(basename) if directory is None else generate_path(basename, directory)
 
-    for key in _get_s3_keys(bucket, key_base_prefixed):
-        download_s3(bucket, key, generate_path(os.path.relpath(key, key_base), directory))
+    if not overwrite and exists(target):
+        return target
 
-    return directory
+    elif path.startswith('s3'):
+        cache = generate_path(basename)
 
+        if overwrite or not exists(cache):
+            bucket, key = get_bucket_and_key(path)
+            download_s3(bucket, key, cache)
 
-def localize_file(path: str, directory: Optional[str] = None, overwrite: bool = True) -> str:
-    if not path.startswith('s3'):
+        if cache != target:
+            shutil.copy(cache, target)
+
+        return target
+
+    elif directory is not None:
+        if path != target:
+            shutil.copy(path, target)
+
+        return target
+
+    else:
         return path
 
-    bucket, key = get_bucket_and_key(path)
-
-    directory = generate_path(os.path.basename(path)) if directory is None else directory
-
-    if not overwrite and os.path.exists(directory):
-        return directory
-
-    download_s3(bucket, key, directory)
-
-    return directory
-
 
 def localize(
         path: str,
-        directory: Optional[str] = None,
+        directory: Optional[str] = os.getenv('DATA_EPHEMERAL_PATH'),
         overwrite: bool = True,
         override: bool = utils.to_bool_or_false(os.getenv('DEVELOPMENT_MODE')),
 ) -> str:
