@@ -89,22 +89,28 @@ class MLFlowModelCheckpoint(ModelCheckpoint):
 
         io.upload_s3(bucket, destination_path, local_file)
 
-    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
-        super().on_train_epoch_end(trainer, pl_module)
+    def has_new_model(self) -> bool:
+        return bool(self.best_model_path) and self.best_model_path != self.mlflow_model_path
 
-        should_save = trainer.current_epoch - self.mlflow_epoch >= self.patience
-        has_new_model = self.best_model_path and self.best_model_path != self.mlflow_model_path
-
-        if should_save and has_new_model:
+    def update_model(self, trainer: Trainer) -> None:
+        if self.has_new_model() and io.exists(self.best_model_path):
             self.log_artifact(self.best_model_path)
-
             self.mlflow_epoch = trainer.current_epoch
             self.mlflow_model_path = self.best_model_path
 
+    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
+        super().on_train_epoch_end(trainer, pl_module)
+
+        if trainer.current_epoch - self.mlflow_epoch >= self.patience:
+            self.update_model(trainer)
+
     def on_train_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
         super().on_train_end(trainer, pl_module)
+        self.update_model(trainer)
 
-        self.log_artifact(self.best_model_path)
+    def on_exception(self, trainer: Trainer, pl_module: LightningModule, exception: BaseException) -> None:
+        super().on_exception(trainer, pl_module, exception)
+        self.update_model(trainer)
 
 
 class StreamPersister:
@@ -122,6 +128,17 @@ class StreamPersister:
         self.file.flush()
 
 
+def add_params(function: Callable) -> Callable:
+    @functools.wraps(function)
+    def wrapper(*args, **kwargs):
+        parameters = parameters_lib.get_script_parameters(function, ignore_keyword_arguments=False)
+        mlflow.log_params(parameters)
+
+        return function(*args, **kwargs)
+
+    return wrapper
+
+
 def log(function: Callable, override: bool = utils.to_bool_or_false(os.getenv('DEVELOPMENT_MODE'))) -> Callable:
     @functools.wraps(function)
     def wrapper(*args, **kwargs):
@@ -131,12 +148,9 @@ def log(function: Callable, override: bool = utils.to_bool_or_false(os.getenv('D
         sys.stdout = StreamPersister(sys.stdout)
         sys.stderr = StreamPersister(sys.stderr)
 
-        logging.basicConfig(force=True)
+        logging.basicConfig(force=True, level=logging.WARNING)
 
         try:
-            parameters = parameters_lib.get_script_parameters(function)
-            mlflow.log_params(parameters)
-
             return function(*args, **kwargs)
 
         finally:
@@ -148,6 +162,10 @@ def log(function: Callable, override: bool = utils.to_bool_or_false(os.getenv('D
     return wrapper
 
 
+def install(function: Callable, override: bool = utils.to_bool_or_false(os.getenv('DEVELOPMENT_MODE'))) -> Callable:
+    return log(add_params(function), override)
+
+
 def get_metrics(run_id: Optional[str] = None) -> Mapping[str, float]:
     run_id = run_id if run_id else mlflow.active_run().info.run_id
     metrics = mlflow.get_run(run_id).data.metrics if run_id else {}
